[
   {'Key': 'Name', 'Value': '{{ cert.name }}'},
   {'Key': 'Environment', 'Value': 'Development'},
   {'Key': 'Product', 'Value': 'Knox'},
   {'Key': 'Service', 'Value': 'ACM'},
   {'Key': 'Team', 'Value': 'certificate-mgmt@8x8.com'},
   {'Key': 'Owner', 'Value': '8x8'}
]
